"""
Combine several CXI files into one, combine datasets when possible
Author: Sergey Bobkov
"""

from shutil import copy2
from typing import List, Dict, Tuple
import numpy as np
from . import cxidata


def combine_files(input_files: List[str], output_file: str) -> None:
    """Combine several CXI files into one, combining datasets that
    have matching shapes.

    Keyword arguments:
    input_files -- list of input CXI files
    output_file -- filename for output CXI file
    """
    if not input_files:
        raise ValueError('Empty input_files list')

    if len(input_files) == 1:
        copy2(input_files[0], output_file)
        return

    image_id_mapping = _create_image_id_mapping(input_files)

    cxidata.create_file(output_file)

    for i in sorted(set(image_id_mapping.values())):
        output_image_id = cxidata.add_image_group(output_file)
        assert i == output_image_id

        # Select input image groups for current output image group
        input_image_groups = [key for key, val in image_id_mapping.items()
                              if val == output_image_id]

        dataset_names = []
        for fname, image_id in input_image_groups:
            dataset_names.extend(cxidata.get_names(fname, image_id))
        dataset_names = set(dataset_names)

        for name in dataset_names:
            dataset_image_groups = [(fname, image_id) for fname, image_id in input_image_groups
                                    if name in cxidata.get_names(fname, image_id)]

            _combine_dataset(name, dataset_image_groups, output_file, output_image_id)


def _combine_dataset(name: str, dataset_image_groups: List[Tuple[str, int]],
                     output_file: str, output_image_id: int) -> None:
    first_fname, first_image_id = dataset_image_groups[0]
    compression = cxidata.get_dataset_compression(first_fname, first_image_id, name)
    chunks = cxidata.get_dataset_chunks(first_fname, first_image_id, name)

    if name in ['mask', 'image_center']:
        data = cxidata.read_dataset(first_fname, first_image_id, name)
        cxidata.save_dataset(output_file, output_image_id, name, data,
                             chunks=chunks, compression=compression)
        return

    dataset_shapes = [cxidata.get_dataset_shape(fname, image_id, name)
                      for fname, image_id in dataset_image_groups]

    ndim = len(dataset_shapes[0])

    if ndim == 0:
        data = [cxidata.read_dataset(fname, image_id, name)
                for fname, image_id in dataset_image_groups]

        if all(d == data[0] for d in data):
            data = data[0]
        else:
            data = np.array(data)
        cxidata.save_dataset(output_file, output_image_id, name, data)
    elif ndim == 1:
        for j, shape in enumerate(dataset_shapes):
            if len(shape) != 1:
                fname, image_id = dataset_image_groups[j]
                raise ValueError("Dataset '{}' have incompatible shapes {} and {}".format(
                    name, dataset_shapes[0], shape))

        data = np.concatenate([cxidata.read_dataset(fname, image_id, name)
                               for fname, image_id in dataset_image_groups])
        cxidata.save_dataset(output_file, output_image_id, name, data)
    else:
        element_shape = dataset_shapes[0][1:]

        lenght = 0
        for j, shape in enumerate(dataset_shapes):
            lenght += shape[0]
            if shape[1:] != element_shape:
                fname, image_id = dataset_image_groups[j]
                raise ValueError("Dataset '{}' have incompatible shapes {} and {}".format(
                    name, dataset_shapes[0], shape))

        chunk_size = int(1e8 // (np.prod(element_shape) + 1) + 1)

        dtype = cxidata.get_dataset_dtype(first_fname, first_image_id, name)
        cxidata.create_dataset(output_file, output_image_id, name,
                               (lenght,) + element_shape, dtype, chunks, compression)

        running_length = 0
        for j, shape in enumerate(dataset_shapes):
            fname, image_id = dataset_image_groups[j]

            for start in range(0, shape[0], chunk_size):
                end = min(start + chunk_size, shape[0])

                data = cxidata.read_dataset(fname, image_id, name, start, end)
                cxidata.update_dataset(output_file, output_image_id, name, data,
                                       running_length + start, running_length + end)

            running_length += shape[0]

        assert running_length == lenght


def _create_image_id_mapping(input_files: List[str]) -> Dict[Tuple[str, int], int]:
    """ Create mapping between input files and image_id in output_file

    Keyword arguments:
    input_files -- list of input CXI files

    Return:
    image_id_mapping -- Dict with mapping (input_file, image_id) -> image_id in output file
    """
    image_keys = []
    image_id_mapping = {}

    for fname in input_files:
        file_image_ids = cxidata.get_image_groups(fname)
        for image_id in file_image_ids:
            dset_names = cxidata.get_names(fname, image_id)

            if 'mask' in dset_names:
                mask = cxidata.read_dataset(fname, image_id, 'mask')
            else:
                mask = None

            if 'image_center' in dset_names:
                image_center = cxidata.read_dataset(fname, image_id, 'image_center')
            else:
                image_center = None

            for i, key in enumerate(image_keys):
                saved_mask, saved_center = key
                if saved_mask is None and mask is None:
                    mask_fit = True
                elif saved_mask is not None and mask is not None and \
                    saved_mask.shape == mask.shape and \
                    (saved_mask == mask).all():
                    mask_fit = True
                else:
                    mask_fit = False

                if saved_center is None and image_center is None:
                    center_fit = True
                elif saved_center is not None and image_center is not None and \
                    saved_center.shape == image_center.shape and \
                    (saved_center == image_center).all():
                    center_fit = True
                else:
                    center_fit = False

                if mask_fit and center_fit:
                    image_id_mapping[(fname, image_id)] = i + 1
                    break
            else:
                image_keys.append((mask, image_center))
                image_id_mapping[(fname, image_id)] = len(image_keys)

    return image_id_mapping




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Refine image center for diffraction images in CXI file by fit of sphere form factor.
Author: Sergey Bobkov
"""

import os
import sys
import shutil
import subprocess
import argparse
import numpy as np
from tqdm import tqdm

from spi_processing import center, cxidata, sphere


def main():
    parser = argparse.ArgumentParser(description="Refine image center for diffraction images by fit of sphere form factor")
    parser.add_argument('files', metavar='FILE', nargs='+', help='Input files')
    parser.add_argument('-o', '--out', dest='output_dir', help="Output directory")
    parser.add_argument('-s', '--same', action='store_true',
                        help="Consider all input files to have the same center position")
    parser.add_argument('-r', '--report', default="estimate_center_report.pdf",
                        help="Report file")
    parser.add_argument('--max-shift', type=int, default=5,
                        help="Maximum allowed shift in pixels")
    parser.add_argument('-R', '--radius', type=int, default=128, help="Maximum radius in pixels")
    parser.add_argument('-w', '--wavelength', type=float, required=True,
                        help="Radiation wavelength (Angstrom)")
    parser.add_argument('-d', '--distance', type=float, required=True,
                        help="Detector distance (meters)")
    parser.add_argument('--pix', dest='pixel', type=float, required=True,
                        help="Pixel size (meters)")
    parser_args = parser.parse_args()

    input_files = parser_args.files
    output_dir = parser_args.output_dir
    same_pos = parser_args.same
    report_path = parser_args.report
    max_shift = parser_args.max_shift
    radius = parser_args.radius
    wavelength = parser_args.wavelength
    distance = parser_args.distance
    pixel_size = parser_args.pixel

    for fname in input_files:
        if not os.path.isfile(fname):
            parser.error("File {} doesn't exist".format(fname))

    if output_dir:
        os.makedirs(output_dir, exist_ok=True)
        sys.stderr.write("Copying to output\n")
        os.makedirs(output_dir, exist_ok=True)
        new_input_files = []
        for fname in tqdm(input_files):
            new_fname = os.path.join(output_dir, os.path.basename(fname))
            shutil.copy(fname, new_fname)
            subprocess.call(["/bin/chmod", "u+w", new_fname])
            new_input_files.append(new_fname)
        input_files = new_input_files

    if same_pos:
        sum_data_stack = []
        mask_stack = []
        sys.stderr.write("Reading files\n")
        for fname in tqdm(input_files):
            sum_data_stack.append(center.compute_sum_image(fname))
            mask_stack.append(center.compute_common_mask(fname))

        sum_data = np.stack(sum_data_stack).sum(axis=0)

        # Get pixels that are good for all datasets
        mask = np.abs(np.stack(mask_stack)).min(axis=0)

        image_id = cxidata.get_image_groups(input_files[0])[0]
        start_center = cxidata.read_dataset(input_files[0], image_id, "image_center")
        if "psd/size" in cxidata.get_names(input_files[0], image_id):
            start_size = cxidata.read_dataset(input_files[0], image_id, "psd/size").mean()
        else:
            start_size = sphere.compute_size(sum_data[1]/10, pixel_size/(wavelength*distance))

        sys.stderr.write("Refining center\n")
        image_center = center.refine_center_by_sphere_pattern(sum_data, mask, wavelength, distance,
                                                              pixel_size, start_center, start_size,
                                                              max_shift, radius)

        sys.stderr.write("Saving result\n")
        for fname in tqdm(input_files):
            center.save(fname, image_center)

        sys.stderr.write("Creating report\n")
        center.generate_report(["Summary image"], [sum_data], [image_center], report_path)
    else:
        sum_data_stack = []
        center_values = []
        sys.stderr.write("Processing files\n")
        for fname in tqdm(input_files):
            sum_data = center.compute_sum_image(fname)
            mask = center.compute_common_mask(fname)

            image_id = cxidata.get_image_groups(fname)[0]
            start_center = cxidata.read_dataset(fname, image_id, "image_center")

            if "psd/size" in cxidata.get_names(fname, image_id):
                start_size = cxidata.read_dataset(fname, image_id, "psd/size").mean()
            else:
                start_size = sphere.compute_size(sum_data[1]/10, pixel_size/(wavelength*distance))

            image_center = center.refine_center_by_sphere_pattern(
                sum_data, mask, wavelength, distance, pixel_size,
                start_center, start_size, max_shift, radius)

            center.save(fname, image_center)
            sum_data_stack.append(sum_data)
            center_values.append(image_center)

        sys.stderr.write("Creating report\n")
        center.generate_report(input_files, sum_data_stack, center_values, report_path)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Combine several CXI files into one, combine datasets when possible
Author: Sergey Bobkov
"""

import os
import argparse

from spi_processing import combine


def main():
    parser = argparse.ArgumentParser(description="Combine several CXI files into one, combine datasets when possible")
    parser.add_argument('files', metavar='FILE', nargs='+', help='Input files')
    parser.add_argument('-o', '--out_file', dest='output_file', required=True, help="Output file")
    parser_args = parser.parse_args()

    input_files = parser_args.files
    output_file = parser_args.output_file

    for fname in input_files:
        if not os.path.exists(fname) or not os.path.isfile(fname):
            parser.error("File {} doesn't exist".format(fname))

    if os.path.exists(output_file):
        parser.error("File {} already exists".format(output_file))

    combine.combine_files(input_files, output_file)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Filter CXI file by dataset values
Author: Sergey Bobkov
"""

import os
import shutil
import tempfile
import argparse
import pandas as pd
from tqdm import tqdm

from spi_processing import combine, filtering, pdfreport


def main():
    parser = argparse.ArgumentParser(description='Filter CXI data by dataset values')
    parser.add_argument('files', metavar='FILE', nargs='+', help='Input files')
    parser.add_argument('-d', '--dset', dest='dset', required=True, help="Filtering dataset")
    parser.add_argument('-o', '--outdir', dest='output_dir', help="Output directory")
    parser.add_argument('--outfile', dest='output_file', help="Output file")
    parser.add_argument('-m', '--min', dest='min_value', type=float, help="Minimum size")
    parser.add_argument('-M', '--max', dest='max_value', type=float, help="Maximum size")
    parser.add_argument('-r', '--report', default='filter_report.pdf', help="Report file")
    parser_args = parser.parse_args()

    input_files = parser_args.files
    dset_name = parser_args.dset
    output_dir = parser_args.output_dir
    output_file = parser_args.output_file
    min_value = parser_args.min_value
    max_value = parser_args.max_value
    report_path = parser_args.report

    for fname in input_files:
        if not os.path.isfile(fname):
            parser.error("File {} doesn't exist".format(fname))

    if output_dir is not None:
        if os.path.exists(output_dir) and not os.path.isdir(output_dir):
            parser.error("{} is not a directory".format(output_dir))
        elif not os.path.exists(output_dir):
            os.makedirs(output_dir)

    if output_file is not None:
        if os.path.exists(output_file):
            parser.error("File {} already exists".format(output_file))

    if report_path is not None:
        report_df = pd.DataFrame(columns=('Filename', 'Total images', 'Selected images'))

    tempdir = tempfile.TemporaryDirectory()
    if output_dir is None:
        output_dir = tempdir.name

    output_files = []
    for fname in tqdm(input_files):
        out_fname = os.path.join(output_dir, os.path.basename(fname))
        tot_n, sel_n = filtering.filter_file(fname, out_fname, dset_name, min_value, max_value)
        output_files.append(out_fname)

        if report_path is not None:
            report_df.loc[len(report_df)] = [os.path.split(fname)[1], tot_n, sel_n]

    if report_path is not None:
        report_df.loc[len(report_df)] = \
            ['Total', report_df.iloc[:, 1].sum(), report_df.iloc[:, 2].sum()]
        rep = pdfreport.ReportWriter(report_path)
        rep.save_table(report_df)
        rep.close()

    if output_file is not None:
        if len(output_files) == 1:
            shutil.move(output_files[0], output_file)
        else:
            combine.combine_files(output_files, output_file)

    tempdir.cleanup()


if __name__ == '__main__':
    main()



"""
Scripts for estimation of image center for diffraction images in CXI file
Author: Sergey Bobkov
"""

import os
from typing import List
import numpy as np
from scipy.optimize import least_squares, brute
from scipy.spatial import distance as sp_distance
import matplotlib.pyplot as plt

from . import cxidata, pdfreport, sphere
from .particle_size import _compute_psd_numba


def save(input_file: str, image_center: np.ndarray) -> None:
    """Save image_center position if CXI file

    Keyword arguments:
    input_file -- name of file
    image_center -- numpy array with center coordinates [x, y, z] in pixels
    """

    dset_name = 'image_center'

    if not isinstance(image_center, np.ndarray) or image_center.shape != (3,):
        raise ValueError("Expected numpy array with 3 elements")

    image_ids = cxidata.get_image_groups(input_file)

    for image_id in image_ids:
        if dset_name in cxidata.get_names(input_file, image_id):
            cxidata.delete_dataset(input_file, image_id, dset_name)

        cxidata.save_dataset(input_file, image_id, dset_name, image_center)


def estimate_center_by_rotate_symmetry(data: np.ndarray, mask: np.ndarray, min_points: int = 20,
                                       relative_distance_lim: float = 0.1, loss: str = 'linear'):
    """Estimate beam position by rotational symmetry
    (equal distance to points with equal intensity)

    Keyword arguments:
    data -- input image
    mask -- mask of good pixels, 0 - good, other values - bad.
    min_points -- minimal number of point with equal intensity to consider that intensity
    relative_distance_lim -- maximum allowed difference of relative distance to points
    loss -- loss parameter for scipy.optimize.least_squares
            ('linear', 'soft_l1', 'huber', 'cauchy', 'arctan')
    """

    if not isinstance(data, np.ndarray) or data.ndim != 2:
        raise ValueError("Expected 2D numpy array for 'data'")

    if not isinstance(mask, np.ndarray) or mask.ndim != 2:
        raise ValueError("Expected 2D numpy array for 'mask'")

    if not np.issubdtype(data.dtype, np.integer):
        raise ValueError("Expected integer dtype for data")

    if mask.shape != data.shape:
        raise ValueError("'mask' shape does not fit with 'data'")

    size_y, size_x = data.shape

    x_range = np.arange(size_x)
    y_range = np.arange(size_y)
    x_coord, y_coord = np.meshgrid(x_range, y_range)

    mask = mask.flatten()
    good_pixels = (mask == 0)
    x_coord = x_coord.flatten()[good_pixels]
    y_coord = y_coord.flatten()[good_pixels]
    intensity = data.flatten()[good_pixels]

    center_stack = []

    vals, counts = np.unique(intensity, return_counts=True)

    for val, count in zip(vals, counts):
        if count < min_points:
            continue

        sel_x = x_coord[intensity == val]
        sel_y = y_coord[intensity == val]
        points = np.vstack([sel_x, sel_y]).T
        try:
            center = _fit_equal_distance(points, loss)
        except ValueError:
            continue

        cdist = sp_distance.cdist([center], points)[0]
        mean_distance = cdist.mean()
        max_diff = np.abs(cdist - mean_distance).max()

        if max_diff > relative_distance_lim*mean_distance or \
                max_diff > relative_distance_lim*max(size_x, size_y):
            continue

        center_stack.append([center[0], center[1], 0])

    if not center_stack:
        raise ValueError("Cannot find suited intensity values")

    center_stack = np.array(center_stack)
    center = center_stack.mean(axis=0)
    return np.round(center, decimals=1)


def refine_center_by_sphere_pattern(data: np.ndarray, mask: np.ndarray,
                                    wavelength: float, distance: float, pixel_size: float,
                                    start_center: np.ndarray, start_size: float,
                                    max_shift: int = 5, radius: int = 128) -> np.ndarray:
    """Search for center position with best fit between image and sphere scattering pattern

    Keyword arguments:
    data -- input image
    mask -- mask of good pixels, 0 - good, other values - bad.
    wavelength -- radiation wavelenght (Angstrom)
    distance -- distance between detector and interaction point (m)
    pixel_size -- size of one detector pixel (m)
    start_center -- position of center to start with (pixels)
    start_size -- starting size a sphere
    max_shift -- maximum difference between strating center and the result, affects performance
    radius -- maximum radius in pixels, that are taken into account

    Return:
    center -- position with best fitting quality
    """

    if not isinstance(data, np.ndarray) or data.ndim != 2:
        raise ValueError("Expected 2D numpy array for 'data'")

    if not isinstance(mask, np.ndarray) or mask.ndim != 2:
        raise ValueError("Expected 2D numpy array for 'mask'")

    if mask.shape != data.shape:
        raise ValueError("'mask' shape does not fit with 'data'")

    freq_step = pixel_size/(wavelength*distance)

    good_data = data[mask == 0]

    def center_fit_qual(center):
        psd_data = _compute_psd_numba(data, mask, center, r_max=radius)
        fit_res = sphere.refine_size(psd_data, start_size, freq_step)
        sphere_pattern = sphere.simulate_pattern(mask.shape, center, fit_res['r_mid'],
                                                 fit_res['scale'], freq_step)

        sphere_pattern = sphere_pattern[mask == 0]
        sphere_pattern *= good_data.mean()/sphere_pattern.mean()
        diff = np.abs(sphere_pattern - good_data).sum()
        return diff

    bounds = [(start_center[0]-max_shift, start_center[0]+max_shift),
              (start_center[1]-max_shift, start_center[1]+max_shift)]
    # result = differential_evolution(center_fit_qual, bounds)
    # center = np.round(result.x, decimals=1)

    center = brute(center_fit_qual, bounds, Ns=(max_shift*10 + 1)) # point every 0.2 pixels
    center = np.round(center, decimals=1)

    return np.array([center[0], center[1], 0])


def compute_sum_image(input_file: str) -> np.ndarray:
    """
    Compute sum of all images in 'data' in CXI file

    Keyword arguments:
    input_file -- input CXI files

    Return:
    sum_image -- 2D array with sum of all images
    """

    sum_image_stack = []
    image_ids = cxidata.get_image_groups(input_file)
    for image_id in image_ids:
        sum_image_stack.append(cxidata.compute_dataset_sum(input_file, image_id, 'data', 0))

    if not sum_image_stack:
        raise ValueError('Empty file: {}'.format(input_file))

    return np.stack(sum_image_stack).sum(axis=0)


def compute_common_mask(input_file: str) -> np.ndarray:
    """
    Compute mask with good pixels in 'mask' in all CXI file

    Keyword arguments:
    input_file -- input CXI file

    Return:
    mask -- 2D array with common mask
    """

    mask_stack = []
    image_ids = cxidata.get_image_groups(input_file)
    for image_id in image_ids:
        mask_stack.append(cxidata.read_dataset(input_file, image_id, 'mask'))

    if not mask_stack:
        raise ValueError('Empty file: {}'.format(input_file))

    return np.abs(np.stack(mask_stack)).min(axis=0)


def normalise_for_estimate(data: np.ndarray) -> np.ndarray:
    """
    Rescale image into 0-100 intensity range and round to integer values

    Keyword arguments:
    data -- input numpy array

    Return:
    result -- rescaled input converted into integer
    """

    max_val = data.max()
    if max_val > 100:
        data = 100 * (data/max_val)
    return np.round(data).astype(int)


def generate_report(input_files: List[str], sum_images: List[np.ndarray],
                    image_center_list: List[np.ndarray], report_path: str):
    """Create PDF report

    Keyword arguments:
    input_files -- paths of input CXI file
    sum_images -- list of sum_data for input files
    image_center_list -- list of image_center values for input files
    report_path -- path for result pdf file
    """

    assert len(input_files) == len(sum_images)
    assert len(input_files) == len(image_center_list)

    rep = pdfreport.ReportWriter(report_path)

    for fname, sum_image, image_center in zip(input_files, sum_images, image_center_list):
        fig = _create_report_fig(sum_image, image_center)
        fig.suptitle(os.path.basename(fname), fontsize=20)
        rep.save_figure(fig)

    rep.close()


def _create_report_fig(sum_data, image_center):
    """Plot figure for PDF report"""
    fig, axes = plt.subplots(figsize=(10, 10), dpi=200)
    pdfreport.plot_image(sum_data, axes)
    axes.plot(image_center[0], image_center[1], '+', color='r', markersize=30, linewidth=0.1)
    axes.add_artist(plt.Circle(image_center, 70, color='r', fill=False))
    axes.add_artist(plt.Circle(image_center, 50, color='r', fill=False))

    return fig


def _find_intersection(base_1, inc_1, base_2, inc_2):
    """Find the intersection point of two lines

    lines are defined as:
        base = (x, y)
        inc = (dx, dy)
        base + inc*t

    Keyword arguments:
    base_1, inc_1 -- first line
    base_2, inc_2 -- second line

    Return:
        res_x, res_y: coordinates of intersection
    """
    # x = x1 + dx1*t
    # y = y1 + dy1*t
    # y = y1 + (x - x1)*dy1/dx1
    # y = y2 + (x - x2)*dy2/dx2
    # (x - x1)*dy1/dx1 - (x - x2)*dy2/dx2 = y2 - y1
    # x*(dy1/dx1 - dy2/dx2) - x1*dy1/dx1 + x2*dy2/dx2 = y2 - y1
    # x = (y2 - y1 + (x1*dy1/dx1 - x2*dy2/dx2))/(dy1/dx1 - dy2/dx2)

    if inc_1[0] == 0:
        inc_1[0] = 1e-10
    if inc_2[0] == 0:
        inc_2[0] = 1e-10

    slope_1 = inc_1[1]/inc_1[0]
    slope_2 = inc_2[1]/inc_2[0]

    if slope_1 == slope_2:
        raise ValueError("Lines do not intersect (they are parallel)")

    res_x = (base_2[1] - base_1[1] + (base_1[0]*slope_1 - base_2[0]*slope_2))/(slope_1 - slope_2)
    res_y = base_1[1] + (res_x - base_1[0])*slope_1

    return res_x, res_y


def _triangulate(point_1, point_2, point_3):
    """Find center of a ring that pass through 3 points

    Keyword arguments:
    point_1 -- first point (x, y)
    point_2 -- second point (x, y)
    point_3 -- third point (x, y)

    Return:
    res_x, res_y -- center of a ring
    """
    # perpendicular bisector
    # y - (y1+y2)/2 = (x - (x1+x2)/2)*(-1*(x2 - x1)/(y2 - y1))
    # x1 = (x1+x2)/2
    # y1 = (y1+y2)/2
    # dx1 = -1*(x2 - x1)
    # dy1 = (y2 - y1)

    # bisector 1-2:
    base_1 = ((point_1[0] + point_2[0])/2, (point_1[1] + point_2[1])/2)
    inc_1 = ((point_2[1] - point_1[1]), -1*(point_2[0] - point_1[0]))
    # bisector 2-3:
    base_2 = ((point_2[0] + point_3[0])/2, (point_2[1] + point_3[1])/2)
    inc_2 = ((point_3[1] - point_2[1]), -1*(point_3[0] - point_2[0]))

    res_x, res_y = _find_intersection(base_1, inc_1, base_2, inc_2)

    return res_x, res_y


def _select_3_utmost_points(points):
    """Select 3 points from the list that are located farthest from each other

    Keyword arguments:
    points -- 2d array with shape (N, 2).
              points[;, 0] - x coorrdinates, points[;, 1] - y coorrdinates

    Return:
    idx1, idx2, idx3 -- indexes of 3 points in points array (first dimension)
    """
    idx1 = 0 # start from first

    # utmost from i1
    idx2 = np.argmax(sp_distance.cdist(points[[idx1]], points)[0])
    # utmost from i2
    idx3 = np.argmax(sp_distance.cdist(points[[idx2]], points)[0])
    # utmost both i2 and i3
    idx1 = np.argmax(np.amin(sp_distance.cdist(points[[idx2, idx3]], points), axis=0))

    return idx1, idx2, idx3


def _fit_equal_distance(points, loss):
    """Find center of a ring that pass as close as possible to all points,
    based on least_squares method from scipy.optimize

    Keyword arguments:
    points -- 2d array with shape (N, 2).
              points[;, 0] - x coorrdinates, points[;, 1] - y coorrdinates
    loss -- loss parameter for scipy.optimize.least_squares
            ('linear', 'soft_l1', 'huber', 'cauchy', 'arctan')

    Return:
    res.x -- coordinates array [x, y]
    """
    idx1, idx2, idx3 = _select_3_utmost_points(points)

    start_x, start_y = _triangulate(points[idx1], points[idx2], points[idx3])

    def fun_resud(center_point):
        cdist = sp_distance.cdist([center_point], points)[0]
        return np.abs(cdist - cdist.mean())

    res = least_squares(fun_resud, [start_x, start_y], loss=loss)

    return res.x



"""
Compute number of photons and litpixels in CXI file
Author: Sergey Bobkov
"""

import numpy as np
from . import cxidata


def compute_photons(filename: str, chunk_size: int = 1000) -> None:
    """Get dataset shape from CXI file

    Keyword arguments:
    filename -- name of file
    chunk_size -- number of frames processed in one chunk
    """

    image_ids = cxidata.get_image_groups(filename)

    for image_id in image_ids:
        data_shape = cxidata.get_dataset_shape(filename, image_id, 'data')
        if len(data_shape) != 3:
            raise ValueError('Data has wrong number of dimensions: {} != 3'.format(len(data_shape)))

        num_frames = data_shape[0]

        num_photons = np.zeros(num_frames)
        litpixels = np.zeros(num_frames)

        for start in range(0, num_frames, chunk_size):
            end = min(start+chunk_size, num_frames)
            chunk_data = cxidata.read_dataset(filename, image_id, 'data', start, end)
            num_photons[start:end] = np.sum(chunk_data, axis=(1, 2))
            litpixels[start:end] = np.sum(chunk_data > 0, axis=(1, 2))

        if 'num_photons' in cxidata.get_names(filename, image_id):
            cxidata.update_dataset(filename, image_id, 'num_photons', num_photons)
        else:
            cxidata.save_dataset(filename, image_id, 'num_photons', num_photons)

        if 'num_litpixels' in cxidata.get_names(filename, image_id):
            cxidata.update_dataset(filename, image_id, 'num_litpixels', litpixels)
        else:
            cxidata.save_dataset(filename, image_id, 'num_litpixels', litpixels)



"""
Scripts for estimation of particle size for diffraction images in CXI file
Author: Sergey Bobkov
"""

from typing import Tuple
import numpy as np
from numba import njit, prange
from scipy import ndimage
from scipy.signal import argrelmin

from . import cxidata, sphere


def compute_psd_data(input_file: str, radius: int = 0, angular_data_ratio: float = 1,
                     chunk_size: int = 100) -> None:
    """Compute Power Spectral Density for images in CXI file and save it in 'psd/data'

    Keyword arguments:
    input_file -- input CXI file
    radius -- outer radius for PSD calculation
    angular_data_ratio -- ratio of angular data that should be considered
    pds_data_name -- dataset name to save result [default: "psd/data"]
    chunk_size -- number of frames processed in one chunk
    """

    pds_data_name = "psd/data"

    image_ids = cxidata.get_image_groups(input_file)
    for image_id in image_ids:
        data_shape = cxidata.get_dataset_shape(input_file, image_id, 'data')
        num_frames = data_shape[0]
        mask = cxidata.read_dataset(input_file, image_id, 'mask')
        image_center = cxidata.read_dataset(input_file, image_id, 'image_center')

        if radius == 0:
            radius = _max_angle_distance(data_shape[1:], image_center)

        if pds_data_name in cxidata.get_names(input_file, image_id):
            cxidata.delete_dataset(input_file, image_id, pds_data_name)

        cxidata.create_dataset(input_file, image_id, pds_data_name, (num_frames, radius), float)

        if angular_data_ratio < 1:
            proj_data = _prepare_polar_projection(mask.shape, radius)
            polar_mask = _apply_polar_projection(proj_data, mask, image_center, cval=1)

        for start in range(0, num_frames, chunk_size):
            end = min(start+chunk_size, num_frames)
            data_chunk = cxidata.read_dataset(input_file, image_id, 'data', start, end)
            psd_buffer = np.zeros((end - start, radius))

            if angular_data_ratio == 1:
                _compute_psd_array_numba(data_chunk, mask,
                                         image_center, radius,
                                         psd_buffer)
            else:
                for i in range(end - start):
                    polar_data = _apply_polar_projection(proj_data, data_chunk[i],
                                                         image_center, cval=0)
                    polar_data[polar_mask] = 0
                    psd_buffer[i] = _compute_psd_polar(polar_data, polar_mask,
                                                       angular_data_ratio)

            cxidata.update_dataset(input_file, image_id, pds_data_name, psd_buffer, start, end)


def estimate_size(input_file: str, wavelength: float, distance: float, pixel_size: float,
                  r_min: int = 0, r_max: int = 0, interp: int = 5,
                  size_min: float = 0, size_max: float = 0, num_size: int = 300) -> None:
    """Estimate size of particles by PSD data in 'psd/data'

    Keyword arguments:
    input_file -- input CXI file
    wavelength -- radiation wavelenght (Angstrom)
    distance -- distance between detector and interaction point (m)
    pixel_size -- size of one detector pixel (m)
    r_min -- inner radius for PSD fitting (pix)
    r_max -- outer radius for PSD fitting (pix)
    interp -- interpolation factor to avoid collapse of sphere form-factor oscilation
    size_min -- minimum particle size (Angstrom)
    size_max -- maximum particle size (Angstrom)
    num_size -- number of tested sizes within size range
    """
    freq_step = pixel_size/(wavelength*distance)

    image_ids = cxidata.get_image_groups(input_file)
    for image_id in image_ids:
        psd_data = cxidata.read_dataset(input_file, image_id, "psd/data")

        names = cxidata.get_names(input_file, image_id)

        for dset_name in ["psd/scale", "psd/fit_diff", "psd/size_range",
                          "psd/size", "psd/size_score"]:
            if dset_name in names:
                cxidata.delete_dataset(input_file, image_id, dset_name)

        if r_min == 0:
            beamstop = np.argmax(psd_data.max(axis=0) > 0)
            r_min = beamstop + 2
        if r_max == 0:
            r_max = psd_data.shape[1]

        if size_min == 0:
            size_min = sphere.compute_size(psd_data.shape[1]//2, freq_step)
        if size_max == 0:
            size_max = sphere.compute_size(6, freq_step)

        fit_result = _estimate_size_psd_sphere(psd_data, freq_step, size_min, size_max,
                                               r_min, r_max, interp, num_size)

        size = np.zeros(len(psd_data))
        size_score = np.ones(len(psd_data))

        for j, frame_fit_diff in enumerate(fit_result['fit_diff']):
            local_min, score = _get_fit_optimum(frame_fit_diff)
            size[j] = fit_result['size_range'][local_min]
            size_score[j] = score

        cxidata.save_dataset(input_file, image_id, "psd/scale", fit_result['scale'])
        cxidata.save_dataset(input_file, image_id, "psd/fit_diff", fit_result['fit_diff'])
        cxidata.save_dataset(input_file, image_id, "psd/size_range", fit_result['size_range'])
        cxidata.save_dataset(input_file, image_id, "psd/size", size)
        cxidata.save_dataset(input_file, image_id, "psd/size_score", size_score)


def correct_size(input_file: str, size_multiplier: float) -> None:
    """Correct size in CXI file by factor of size_multiplier

    Keyword arguments:
    input_file -- input CXI file
    size_multiplier -- correction factor
    """

    name = "psd/size"

    image_ids = cxidata.get_image_groups(input_file)
    for image_id in image_ids:
        size = cxidata.read_dataset(input_file, image_id, name)
        size *= size_multiplier
        cxidata.update_dataset(input_file, image_id, name, size)


def _max_angle_distance(frame_shape: Tuple[int, int], center: np.ndarray) -> int:
    """Calculate distance from center to farest angle for frame

    Keyword arguments:
    frame_shape -- frame shape (size_y, size_x)
    center -- center coordinates (x, y)

    Return:
    distance -- integer distance value
    """

    max_x_distance = max(center[0], frame_shape[1] - center[0])
    max_y_distance = max(center[1], frame_shape[0] - center[1])
    distance = np.sqrt(max_x_distance**2 + max_y_distance**2)
    return int(np.floor(distance))


@njit(fastmath=True)
def _compute_psd_numba(image_data, mask_data, image_center, r_max):
    psd = np.zeros(r_max)
    work_pixels = np.zeros(r_max)

    start_x = int(np.round(image_center[0]) - (r_max + 1))
    end_x = int(np.round(image_center[0]) + (r_max + 1))

    start_y = int(np.round(image_center[1]) - (r_max + 1))
    end_y = int(np.round(image_center[1]) + (r_max + 1))

    for i in range(start_y, end_y + 1):
        y_dist = i - image_center[1]
        for j in range(start_x, end_x + 1):
            x_dist = j - image_center[0]

            r_value = int(np.round(np.sqrt(x_dist**2 + y_dist**2)))

            if r_value >= r_max:
                continue

            if 0 <= j < image_data.shape[1] and 0 <= i < image_data.shape[0]:
                if mask_data[i, j] == 0:
                    work_pixels[r_value] += 1
                    psd[r_value] += image_data[i, j]

    for i in range(r_max):
        if work_pixels[i] > 0:
            psd[i] /= work_pixels[i]

    return psd


@njit(parallel=True)
def _compute_psd_array_numba(image_data, mask_data, image_center, r_max, out_psd):
    num_frames = len(image_data)
    for i in prange(num_frames):    # pylint: disable=E1133
        out_psd[i] = _compute_psd_numba(image_data[i], mask_data, image_center, r_max)


def _prepare_polar_projection(data_shape, r_max, size_r=0, size_angle=500):
    """Create proj_data that can be used to speed-up multiple polar projections

    Keyword arguments:
    data_shape -- shape of input data
    limit -- outer limit for image: 'corner', 'edge' or distance in pixels from center
    nr -- radial resolution of result, by default equal distance from center in pixels
    nt -- angular resolution of result, by default 500 angles.

    Return:
    proj_data -- dict to input in _apply_polar_projection()
    """
    size_y, size_x = data_shape[:2]

    r_min = 0
    t_max = np.pi
    t_min = -np.pi

    if size_r == 0:
        size_r = int(r_max)

    r_i = np.linspace(r_min, r_max, size_r, endpoint=False)
    theta_i = np.linspace(t_min, t_max, size_angle, endpoint=False)
    theta_grid, r_grid = np.meshgrid(theta_i, r_i)

    x_grid = r_grid * np.cos(theta_grid)
    y_grid = r_grid * np.sin(theta_grid)

    x_grid, y_grid = x_grid.flatten(), y_grid.flatten()
    coords = np.vstack((y_grid, x_grid))

    proj_data = {'coords': coords,
                 'size_x': size_x,
                 'size_y': size_y,
                 'size_r': size_r,
                 'size_angle': size_angle}

    return proj_data


def _apply_polar_projection(proj_data, image_data, image_center=None, cval=0):
    """Apply polar projection by proj_data dict to speed-up multiple polar projections

    Keyword arguments:
    proj_data -- output of _prepare_polar_projection()
    image_data -- input 2d image
    image_center -- center position of polar coordinate system, in pixels
    cval -- default value to fill missed areas

    Return:
    result -- 2d image in polar coordinates with shape = (size_r, size_angle)
    """
    assert isinstance(proj_data, dict)

    if image_center is None:
        image_center = (proj_data['size_x']/2.0-0.5, proj_data['size_y']/2.0-0.5)

    coords = proj_data['coords'].copy()

    coords[0] += image_center[1] # Y coords
    coords[1] += image_center[0] # X coords

    mapped_data = ndimage.map_coordinates(image_data, coords, order=0, cval=cval)

    return mapped_data.reshape((proj_data['size_r'], proj_data['size_angle']))


def _compute_psd_polar(polar_data, polar_mask, angular_data_ratio):
    angular_sum = polar_data.sum(axis=0)
    angle_filter = (angular_sum > angular_sum.max() * (1 - angular_data_ratio))

    psd = polar_data[:, angle_filter].sum(axis=1).astype(np.float)
    psd_norm = (polar_mask[:, angle_filter] == 0).sum(axis=1)
    psd[psd_norm == 0] = 0
    psd[psd_norm > 0] /= psd_norm[psd_norm > 0]

    return psd


def _estimate_size_psd_sphere(psd_data, freq_step, size_min, size_max,
                              r_min, r_max, interp, num_size):
    """Estimate particle size by fit of PSD data with spherical form factor

    Keyword arguments:
    psd -- input PSD data
    size_min -- minimum allowed particle size
    size_max -- maximum allowed particle size
    freq_step -- frequency increment for 1 pixex, equal to pixel_size/(wavelength*distance)
    r_min -- minimum radius (index of PSD array) used to fit
    r_max -- maximum radius (index of PSD array) used to fit
    interp -- interpolation factor to avoid collapse of low period oscilation
    num_size -- number of tested sizes within size range

    Return:
    result - dictionary with fitting result
             {"fit_diff", "size_range", "scale", "fit_x", "fit_base"}
    """

    num_frames = psd_data.shape[0]
    x_vals = np.arange(r_min, r_max)
    x_interp = np.arange(r_min, r_max, 1/interp)

    frequency_interp = x_interp*freq_step

    size_range = np.linspace(size_min, size_max, num_size)
    # r_mid_range = size_range*(1+np.sqrt(5))/(2*np.sqrt(10+2*np.sqrt(5)))

    fit_base = np.zeros((num_size, frequency_interp.size))

    for i in range(num_size):
        fit = sphere.form_factor(frequency_interp, size_range[i]/2, 1)
        fit_base[i] = fit/fit.mean()

    fit_diff = np.zeros((num_frames, num_size))
    scale = np.zeros((num_frames,))

    for i in range(num_frames):
        data_to_fit = np.interp(x_interp, x_vals, psd_data[i, r_min:r_max])
        scale[i] = data_to_fit.mean()
        if scale[i] > 0:
            data_to_fit /= scale[i]
            fit_diff[i] = np.abs(fit_base - data_to_fit).sum(axis=1)
        else:
            # for 0 vector we set best fit at minimal size value
            fit_diff[i] = 2
            fit_diff[i, 0] = 1

    return {'fit_diff':fit_diff, 'size_range':size_range, 'scale':scale,\
            'fit_x':x_interp, 'fit_base':fit_base}


def _get_fit_optimum(frame_fit_diff: np.ndarray) -> Tuple[int, float]:
    """ Get index of optimum fit and calculate fidelity score

    Keyword arguments:
    frame_fit_diff -- 1D array with fit values for one frame against size range

    Return:
    index -- index of frame_fit_diff with best fit
    score -- calculated fidelity score
    """
    local_min = argrelmin(frame_fit_diff, order=5, mode='wrap')[0]
    min_diff = frame_fit_diff[local_min]
    local_min = local_min[np.argsort(min_diff)]
    if local_min.size > 1:
        score = frame_fit_diff[local_min[1]] / frame_fit_diff[local_min[0]]
    else:
        score = 1

    return local_min[0], score



"""
Filter CXI file by dataset values
Author: Sergey Bobkov
"""

import os
from typing import Optional, Tuple
import numpy as np
from . import cxidata


def filter_file(input_file: str, output_file: str, filter_dset_name: str,
                min_value: Optional[float], max_value: Optional[float]) -> Tuple[int, int]:
    """Filter CXI file by filter_dset_name. Filter selection is: min <= values <= max.
    Selection is applied to all datasets with same length as filter dataset.

    Keyword arguments:
    input_file -- input CXI files
    output_file -- output CXI file
    filter_dset_name -- path to filter dataset (relative to image_N/)
    min_value -- minimal selected value. If None - not applied.
    max_value -- maximum selected value. If None - not applied.

    Return:
    total_num -- total number of elements in file
    selected_num -- number of selected elements
    """

    image_ids = cxidata.get_image_groups(input_file)

    if not image_ids:
        raise ValueError('No data in CXI file: {}'.format(input_file))

    cxidata.create_file(output_file)

    total_num = 0
    selected_num = 0

    for image_id in image_ids:
        names = cxidata.get_names(input_file, image_id)

        if filter_dset_name not in names:
            raise ValueError("Dataset '{}' does not exist in image {} file {}".format(
                filter_dset_name, image_id, input_file))

        filter_data = cxidata.read_dataset(input_file, image_id, filter_dset_name)

        if filter_data.ndim != 1:
            raise ValueError("Filter dataset should have 1D shape, found: {}".format(
                filter_data.shape))

        filter_len = filter_data.shape[0]
        selection = np.ones(filter_len, dtype=np.bool)

        if min_value is not None:
            selection = np.logical_and(selection, filter_data >= min_value)

        if max_value is not None:
            selection = np.logical_and(selection, filter_data <= max_value)

        total_num += filter_len
        selected_num += selection.sum()

        if selection.sum() == 0:
            continue

        output_image_id = cxidata.add_image_group(output_file)

        _filter_image(input_file, image_id, output_file, output_image_id, selection)

    if selected_num == 0:
        os.remove(output_file)

    return total_num, selected_num


def _filter_image(input_file: str, input_image_id: int, output_file: str, output_image_id: int,
                  selection: 'np.ndarray[np.bool]') -> None:
    for name in cxidata.get_names(input_file, input_image_id):
        shape = cxidata.get_dataset_shape(input_file, input_image_id, name)
        compression = cxidata.get_dataset_compression(input_file, input_image_id, name)
        chunks = cxidata.get_dataset_chunks(input_file, input_image_id, name)
        dtype = cxidata.get_dataset_dtype(input_file, input_image_id, name)

        ndim = len(shape)
        chunk_size = int(1e8 // (np.prod(shape[1:]) + 1) + 1)

        if ndim == 0:
            data = cxidata.read_dataset(input_file, input_image_id, name)
            cxidata.save_dataset(output_file, output_image_id, name, data)
        elif shape[0] == len(selection):
            cxidata.create_dataset(output_file, output_image_id, name,
                                   (selection.sum(),) + shape[1:], dtype, chunks, compression)

            for start in range(0, shape[0], chunk_size):
                end = min(start + chunk_size, shape[0])

                if selection[start:end].sum() == 0:
                    continue

                data = cxidata.read_dataset(input_file, input_image_id, name, start, end)
                data = data[selection[start:end]]

                out_start = selection[:start].sum()
                out_end = selection[:end].sum()
                cxidata.update_dataset(output_file, output_image_id, name, data,
                                       out_start, out_end)
        else:

            cxidata.create_dataset(output_file, output_image_id, name, shape, dtype,
                                   chunks, compression)

            for start in range(0, shape[0], chunk_size):
                end = min(start + chunk_size, shape[0])

                data = cxidata.read_dataset(input_file, input_image_id, name, start, end)
                cxidata.update_dataset(output_file, output_image_id, name, data, start, end)



"""
Library for CXI file operations
Author: Sergey Bobkov
"""

import os
from typing import List, Optional, Union, Tuple
import errno
import numpy as np
import h5py


def create_file(filename: str, rewrite: bool = False) -> None:
    """Create empty CXI file

    Keyword arguments:
    filename -- name of file
    rewrite -- rewrite content if file exists
    """
    if not rewrite and os.path.exists(filename):
        raise FileExistsError(errno.EEXIST, os.strerror(errno.EEXIST), filename)

    with h5py.File(filename, 'w') as h5file:
        h5file['cxi_version'] = 160
        h5file.create_group('entry_1')


def add_image_group(filename: str) -> int:
    """Add image group to CXI file

    Keyword arguments:
    filename -- name of file

    Return:
    image_id -- id of created group
    """

    if not os.path.exists(filename):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)

    existing_ids = get_image_groups(filename)

    image_id = 1
    while image_id in existing_ids:
        image_id += 1

    with h5py.File(filename, 'a') as h5file:
        entry = h5file['entry_1']
        entry.create_group('image_{}'.format(image_id))
        entry['data_{}'.format(image_id)] = h5py.SoftLink('/entry_1/image_{}'.format(image_id))

    return image_id


def get_image_groups(filename: str) -> List[int]:
    """Get IDs of image groups in CXI file

    Keyword arguments:
    filename -- name of file

    Return:
    ids -- IDs of image groups
    """

    if not os.path.exists(filename):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)

    with h5py.File(filename, 'r') as h5file:
        entry = h5file['entry_1']
        ids = [int(key[6:]) for key in entry.keys()
               if key.startswith('image_') and key[6:].isdigit()]

    return ids


def get_names(filename: str, image_id: int) -> List[str]:
    """Get names of all datasets in image group in CXI file

    Keyword arguments:
    filename -- name of file
    image_id -- id of image group

    Return:
    names -- List of dataset names
    """

    if not os.path.exists(filename):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)

    with h5py.File(filename, 'r') as h5file:
        image = h5file['entry_1/image_{}'.format(image_id)]

        return _get_group_content(image)


def _get_group_content(h5group: h5py.Group) -> List[str]:
    names = []
    for key in h5group.keys():
        if isinstance(h5group.get(key, getlink=True), h5py.SoftLink):
            raise ValueError('SoftLink detected, exit due to undefined behaviour')

        item = h5group[key]
        if isinstance(item, h5py.Dataset):
            names.append(key)
        elif isinstance(item, h5py.Group):
            for name in _get_group_content(item):
                names.append(key + '/' + name)

    return names


def create_dataset(filename: str, image_id: int, name: str, shape: Tuple[int, ...],
                   dtype: type, chunks: Tuple[int, ...] = None, compression: str = None) -> None:
    """Save dataset to CXI file

    Keyword arguments:
    filename -- name of file
    image_id -- id of image group
    name -- dataset name
    shape -- array shape
    dtype -- array data type
    chunks -- hdf chunks for created dataset
    compression -- use compression for dataset
    """

    if not os.path.exists(filename):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)

    with h5py.File(filename, 'a') as h5file:
        image = h5file['entry_1/image_{}'.format(image_id)]

        if name in image:
            raise ValueError('Dataset {} already exists in image {}'.format(name, image_id))

        image.create_dataset(name, shape=shape, dtype=dtype, chunks=chunks, compression=compression)


def delete_dataset(filename: str, image_id: int, name: str) -> None:
    """Delete dataset from CXI file

    Keyword arguments:
    filename -- name of file
    image_id -- id of image group
    name -- dataset name
    """

    if not os.path.exists(filename):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)

    with h5py.File(filename, 'a') as h5file:
        image = h5file['entry_1/image_{}'.format(image_id)]

        if name not in image:
            raise ValueError('Dataset {} does not exists in image {}'.format(name, image_id))

        del image[name]


def save_dataset(filename: str, image_id: int, name: str, data: Union[np.ndarray, str, List],
                 chunks: Tuple[int] = None, compression: str = None) -> None:
    """Save dataset to CXI file

    Keyword arguments:
    filename -- name of file
    image_id -- id of image group
    name -- dataset name
    data -- array with data to save
    chunks -- hdf chunks for created dataset
    compression -- use compression for dataset
    """

    if not os.path.exists(filename):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)

    with h5py.File(filename, 'a') as h5file:
        image = h5file['entry_1/image_{}'.format(image_id)]

        if name in image:
            raise ValueError('Dataset {} already exists in image {}'.format(name, image_id))

        image.create_dataset(name, data=data, chunks=chunks, compression=compression)


def read_dataset(filename: str, image_id: int, name: str,
                 chunk_start: Optional[int] = None, chunk_end: Optional[int] = None) -> np.ndarray:
    """Read dataset from CXI file

    Keyword arguments:
    filename -- name of file
    image_id -- id of image group
    name -- dataset name
    chunk_start -- read chunk of data starting from chunk_start
    chunk_end -- read chunk of data ends at chunk_end

    Return:
    data -- dataset content
    """

    if not os.path.exists(filename):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)

    with h5py.File(filename, 'r') as h5file:
        image = h5file['entry_1/image_{}'.format(image_id)]

        if name not in image:
            raise ValueError('Dataset {} does not exists in image {}'.format(name, image_id))

        if not image[name].shape:
            return image[name][()]

        num_frames = image[name].shape[0]

        if chunk_start is None:
            chunk_start = 0

        if chunk_end is None:
            chunk_end = num_frames

        if not 0 <= chunk_start < num_frames:
            raise ValueError('Chunk_start should be >= 0 and < {}'.format(num_frames))

        if not 0 <= chunk_end <= num_frames:
            raise ValueError('Chunk_end should be >= 0 and <= {}'.format(num_frames))

        if chunk_start >= chunk_end:
            raise ValueError('Chunk_start should be < chunk_end')

        return image[name][chunk_start:chunk_end]


def update_dataset(filename: str, image_id: int, name: str, data: np.ndarray,
                   chunk_start: Optional[int] = None, chunk_end: Optional[int] = None) -> None:
    """Save dataset to CXI file

    Keyword arguments:
    filename -- name of file
    image_id -- id of image group
    name -- dataset name
    data -- array with data to save
    chunk_start -- update chunk of data starting from chunk_start
    chunk_end -- update chunk of data ends at chunk_end
    """

    if not os.path.exists(filename):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)

    if chunk_start is not None and chunk_start < 0:
        raise ValueError('Chunk_start should be >= 0')

    if chunk_end is not None and chunk_end < 0:
        raise ValueError('Chunk_end should be >= 0')

    with h5py.File(filename, 'a') as h5file:
        image = h5file['entry_1/image_{}'.format(image_id)]

        if name not in image:
            raise ValueError('Dataset {} does not exists in image {}'.format(name, image_id))

        if chunk_start is None:
            chunk_start = 0
        if chunk_end is None:
            chunk_end = len(image[name])

        update_shape = (chunk_end - chunk_start, ) + image[name].shape[1:]

        if update_shape != data.shape:
            raise ValueError('Data shape does not fit update chunk: {} != {}'.format(data.shape,
                                                                                     update_shape))

        image[name][chunk_start:chunk_end] = data


def get_dataset_shape(filename: str, image_id: int, name: str) -> Tuple[int]:
    """Get dataset shape from CXI file

    Keyword arguments:
    filename -- name of file
    image_id -- id of image group
    name -- dataset name

    Return:
    shape -- tuple with dataset shape
    """

    if not os.path.exists(filename):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)

    with h5py.File(filename, 'r') as h5file:
        image = h5file['entry_1/image_{}'.format(image_id)]

        if name not in image:
            raise ValueError('Dataset {} does not exists in image {}'.format(name, image_id))

        return image[name].shape


def get_dataset_dtype(filename: str, image_id: int, name: str) -> type:
    """Get dataset shape from CXI file

    Keyword arguments:
    filename -- name of file
    image_id -- id of image group
    name -- dataset name

    Return:
    type -- array data type
    """

    if not os.path.exists(filename):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)

    with h5py.File(filename, 'r') as h5file:
        image = h5file['entry_1/image_{}'.format(image_id)]

        if name not in image:
            raise ValueError('Dataset {} does not exists in image {}'.format(name, image_id))

        return image[name].dtype


def get_dataset_chunks(filename: str, image_id: int, name: str) -> Tuple[int, ...]:
    """Get chunks option for dataset in CXI file

    Keyword arguments:
    filename -- name of file
    image_id -- id of image group
    name -- dataset name

    Return:
    chunks -- dataset chunks
    """

    if not os.path.exists(filename):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)

    with h5py.File(filename, 'r') as h5file:
        image = h5file['entry_1/image_{}'.format(image_id)]

        if name not in image:
            raise ValueError('Dataset {} does not exists in image {}'.format(name, image_id))

        return image[name].chunks


def get_dataset_compression(filename: str, image_id: int, name: str) -> str:
    """Get chunks option for dataset in CXI file

    Keyword arguments:
    filename -- name of file
    image_id -- id of image group
    name -- dataset name

    Return:
    compression -- dataset compression
    """

    if not os.path.exists(filename):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)

    with h5py.File(filename, 'r') as h5file:
        image = h5file['entry_1/image_{}'.format(image_id)]

        if name not in image:
            raise ValueError('Dataset {} does not exists in image {}'.format(name, image_id))

        return image[name].compression


def compute_dataset_sum(filename: str, image_id: int, name: str,
                        axis: Union[Tuple[int], int, None] = None,
                        chunk_size: int = 100) -> np.ndarray:
    """Get sum of dataset from CXI file

    Keyword arguments:
    filename -- name of file
    image_id -- id of image group
    name -- dataset name
    axis -- axis to compute sum
    chunk_size -- number of frames in one chunk

    Return:
    result -- array with computed sum
    """

    if not os.path.exists(filename):
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), filename)

    with h5py.File(filename, 'r') as h5file:
        image = h5file['entry_1/image_{}'.format(image_id)]

        if name not in image:
            raise ValueError('Dataset {} does not exists in image {}'.format(name, image_id))

        sum_stack = []
        num_frames = image[name].shape[0]
        for start in range(0, num_frames, chunk_size):
            end = min(start+chunk_size, num_frames)

            sum_stack.append(np.sum(image[name][start:end], axis=axis))

        return np.stack(sum_stack).sum(axis=0)



"""
Correct unwanted background scattering on diffraction images
Author: Sergey Bobkov
"""

import os
from typing import Optional, List
import numpy as np
from numba import njit
from scipy.stats import norm as sp_norm
from scipy.optimize import curve_fit
import matplotlib.pyplot as plt

from . import cxidata, pdfreport


def correct_background(input_file: str, max_background: int = 50,
                       overflow_level: Optional[int] = None, chunk_size: int = 100):
    """Correct background in input CXI file

    Keyword arguments:
    input_file -- CXI file
    max_background -- maximum background intensity
    overflow_level -- maximum intensity when detector is still linear
    chunk_size -- number of frames processed in one chunk
    """

    image_ids = cxidata.get_image_groups(input_file)

    if not image_ids:
        raise ValueError('No data in CXI file: {}'.format(input_file))

    _, size_y, size_x = cxidata.get_dataset_shape(input_file, image_ids[0], 'data')
    data_histogram = np.zeros((max_background, size_y, size_x), dtype=np.int32)

    for image_id in image_ids:
        data_shape = cxidata.get_dataset_shape(input_file, image_id, 'data')

        if data_shape[1:] != (size_y, size_x):
            raise ValueError('Data shape is different for image groups 1 and {}'.format(image_id))

        num_frames = data_shape[0]

        for start in range(0, num_frames, chunk_size):
            end = min(start+chunk_size, num_frames)
            data_chunk = cxidata.read_dataset(input_file, image_id, 'data', start, end)
            _add_chunk_to_histogram(data_chunk.astype(np.int32), data_histogram)

    background = _estimate_background(data_histogram)

    for image_id in image_ids:
        num_frames = cxidata.get_dataset_shape(input_file, image_id, 'data')[0]
        mask = cxidata.read_dataset(input_file, image_id, 'mask')

        for start in range(0, num_frames, chunk_size):
            end = min(start+chunk_size, num_frames)
            data_chunk = cxidata.read_dataset(input_file, image_id, 'data', start, end)

            new_data_chunk = data_chunk - background
            overflow_mask = (data_chunk > overflow_level)
            new_data_chunk[overflow_mask] = data_chunk[overflow_mask]
            new_data_chunk[np.logical_or(mask != 0, new_data_chunk < 0)] = 0

            cxidata.update_dataset(input_file, image_id, 'data', new_data_chunk, start, end)


@njit
def _add_chunk_to_histogram(data_chunk: np.ndarray, histogram: np.ndarray):
    """Compute histogram for chunk of data and add result to existing histogram

    Keyword arguments:
    data_chunk -- input chunk of data (n_frame, y, x)
    histogram -- existing histogram
    """
    max_val, size_y, size_x = histogram.shape
    n_frames = len(data_chunk)

    for i in range(n_frames):
        for y_coord in range(size_y):
            for x_coord in range(size_x):
                val = data_chunk[i, y_coord, x_coord]
                if val < max_val:
                    histogram[val, y_coord, x_coord] += 1


def _normal_pdf(x_val: float, mu_val: float, amplitude_scale: float, width_scale: float):
    return amplitude_scale*sp_norm(mu_val, width_scale).pdf(x_val)


def _estimate_background(data_histogram: np.ndarray):
    max_value, size_y, size_x = data_histogram.shape
    result = np.zeros((size_y, size_x), dtype=int)
    fit_range = np.arange(max_value)

    most_frequent_values = np.argmax(data_histogram, axis=0)

    # Save coordinates where most frequent values > 0
    y_vals, x_vals = np.nonzero(most_frequent_values)

    for y_val, x_val in zip(y_vals, x_vals):
        pixel_hist = data_histogram[:, y_val, x_val]
        pix_most_freq = most_frequent_values[y_val, x_val]

        # Fit gaussian
        start_mean = pix_most_freq
        start_std = pix_most_freq/5
        start_scale = pixel_hist[pix_most_freq]/_normal_pdf(pix_most_freq, start_mean, 1, start_std)
        pstart = [start_mean, start_scale, start_std]

        try:
            popt, _ = curve_fit(_normal_pdf, fit_range, pixel_hist, p0=pstart)
        except RuntimeError:    # Estimation fails
            popt = pstart

        result[y_val, x_val] = max(np.floor(popt[0]), 0) # Background cannot be negative

    return result


def compute_mean_data(filename: str):
    """Compute mean image for CXI file

    Keyword arguments:
    filename -- CXI file

    Return:
    mean_image -- mean image
    """

    image_ids = cxidata.get_image_groups(filename)

    sum_stack = []
    num_frames_stack = []

    for image_id in image_ids:
        data_shape = cxidata.get_dataset_shape(filename, image_id, 'data')
        sum_data = cxidata.compute_dataset_sum(filename, image_id, 'data', axis=0)
        num_frames_stack.append(data_shape[0])
        sum_stack.append(sum_data)

    mean_image = np.stack(sum_stack).sum(axis=0) / np.array(num_frames_stack).sum()

    return mean_image


def _create_report_fig(old_mean_data: np.ndarray, new_mean_data: np.ndarray):
    """Plot figure for PDF report"""
    background = old_mean_data - new_mean_data

    fig, axis = plt.subplots(1, 3, figsize=(15, 10), dpi=200)
    pdfreport.plot_image(old_mean_data, axis[0], logscale=False)
    axis[0].set_title('Original mean image')
    pdfreport.plot_image(new_mean_data, axis[1], logscale=False)
    axis[1].set_title('Corrected mean image')
    pdfreport.plot_image(background, axis[2], logscale=False)
    axis[2].set_title('Estimated background')

    return fig


def generate_report(input_files: List[str], old_mean_list: List[np.ndarray],
                    new_mean_list: List[np.ndarray], report_path: str) -> None:
    """Create PDF report

    Keyword arguments:
    input_files -- paths of input CXI file
    old_mean_list -- list of mean data for files before correction
    new_mean_list -- list of mean data for files after correction
    report_path -- path for result pdf file
    """
    assert len(input_files) == len(old_mean_list)
    assert len(input_files) == len(new_mean_list)

    rep = pdfreport.ReportWriter(report_path)

    for fname, old_mean_data, new_mean_data in zip(input_files, old_mean_list, new_mean_list):
        fig = _create_report_fig(old_mean_data, new_mean_data)
        fig.suptitle(os.path.basename(fname), fontsize=20)
        rep.save_figure(fig)

    rep.close()



"""
Library PDF reports
Author: Sergey Bobkov
"""

import datetime
from typing import Optional
import pandas as pd
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib.backends.backend_pdf import PdfPages
import matplotlib.colors as colors
from mpl_toolkits.axes_grid1 import make_axes_locatable

mpl.rcParams['image.cmap'] = 'viridis'


class ReportWriter():
    """Produce PDF reports with tables and matplotlib figures"""
    def __init__(self, filename: str):
        self._create_file(filename)

    def _create_file(self, filename: str):
        self.pdf_pages = PdfPages(filename)
        pdf_dict = self.pdf_pages.infodict()
        pdf_dict['CreationDate'] = datetime.datetime.today()
        pdf_dict['ModDate'] = datetime.datetime.today()

    def save_table(self, data_frame: pd.DataFrame):
        """Add a page to PDF with table

        Keyword arguments:
        data_frame -- pandas dataframe to be saved
        """
        # Compute optimal columns width
        content_df = data_frame.copy()
        content_df.loc[len(content_df)] = data_frame.columns.values
        col_widhts = np.array([content_df.loc[:, c].apply(str).apply(len).max()\
                               for c in content_df.columns.values], dtype=np.float64)
        col_widhts *= 1/col_widhts.sum()

        fig, axes = plt.subplots(figsize=(10, len(content_df)/4 + 1))
        axes.axis('off')
        axes.axis('tight')

        axes.table(cellText=data_frame.values, colLabels=data_frame.columns, loc='center',
                   colWidths=col_widhts)

        fig.tight_layout()

        self.pdf_pages.savefig()

    def save_figure(self, figure: plt.Figure):
        """Add a page to PDF with matplotlib figure

        Keyword arguments:
        figure -- figure to be saved
        """

        self.pdf_pages.savefig(figure)
        plt.close(figure)


    def close(self):
        """Closes PDF document"""
        self.pdf_pages.close()


def plot_image(image: np.ndarray, axis: plt.Axes, logscale: bool = True,
               colorbar: bool = True, vmin: Optional[float] = None, vmax: Optional[float] = None,
               **kwargs):
    """Plot image to axis with optional logscale and colorbar

    Keyword arguments:
    image -- 2D array with data to show
    axis -- matplotlib axis to plot
    logscale -- use logscale
    colorbar -- add colorbar to image
    vmin -- minimal value of colormap
    vmax -- miximum value of colormap
    kwargs -- arguments for imshow
    """
    if vmin is None and logscale:
        vmin = max(1, np.amin(image))
    elif vmin is None:
        vmin = np.amin(image)

    if vmax is None:
        vmax = np.amax(image)

    if logscale:
        norm = colors.LogNorm(vmin=vmin, vmax=vmax)
    else:
        norm = colors.Normalize(vmin=vmin, vmax=vmax)

    im_handle = axis.imshow(image, norm=norm, **kwargs)
    axis.set_xticks([])
    axis.set_yticks([])

    if colorbar:
        divider = make_axes_locatable(axis)
        cb_axis = divider.append_axes("right", size="5%", pad=0.05)
        plt.colorbar(im_handle, cax=cb_axis)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Plot histogram for dataset in CXI file
Author: Sergey Bobkov
"""

import os
import sys
import argparse
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle
from tqdm import tqdm

from spi_processing import cxidata, pdfreport


def plot_histogram(data, dset_name, bins=100, x_lim=None, select=None):
    fig, ax = plt.subplots(figsize=(10, 6), dpi=200)

    ax.hist(data, bins=bins, range=x_lim)
    ax.set_yscale('log')

    if select is not None:
        x_start = select[0]
        width = select[1] - select[0]
        rect = Rectangle((x_start,0), width, ax.get_ylim()[1],
                         hatch='///', color='C2', lw=1, fill=False)
        ax.add_artist(rect)

    ax.set_ylabel('Number of images')
    ax.set_xlabel(dset_name + ' values')
    return fig


def parse_range(range_str):
    if isinstance(range_str, str) and len(range_str.split(':')) == 2:
        return [int(val) for val in range_str.split(':')]
    else:
        return None


def main():
    parser = argparse.ArgumentParser(description='Plot histogram of CXI dataset')
    parser.add_argument('files', metavar='FILE', nargs='+', help='Input files')
    parser.add_argument('-d', '--dset', required=True, help="Dataset")
    parser.add_argument('-r', '--range', metavar='START:END', type=parse_range,
                        help="Histogram range")
    parser.add_argument('-s', '--select', metavar='START:END', type=parse_range,
                        help="Selection range")
    parser.add_argument('-b', '--bins', type=int, default=100, help="Number of bins")
    parser.add_argument('-o', '--out_file', dest='output_file', default="data_histogram.pdf",
                        help="Output file")
    parser_args = parser.parse_args()

    input_files = parser_args.files
    dset_name = parser_args.dset
    hist_range = parser_args.range
    sel_range = parser_args.select
    bins = parser_args.bins
    output_file = parser_args.output_file

    for fname in input_files:
        if not os.path.isfile(fname):
            parser.error("File {} doesn't exist".format(fname))

    if os.path.exists(output_file):
        parser.error("File {} already exists".format(output_file))

    data_stack = []
    sys.stderr.write("Reading files\n")
    for fname in tqdm(input_files):
        image_ids = cxidata.get_image_groups(fname)
        for image_id in image_ids:
            data = cxidata.read_dataset(fname, image_id, dset_name)
            data_stack.append(data.flatten())

    data = np.concatenate(data_stack)

    sys.stderr.write("Creating histogram\n")
    rep = pdfreport.ReportWriter(output_file)
    rep.save_figure(plot_histogram(data, dset_name, bins, hist_range, sel_range))
    rep.close()

if __name__ == '__main__':
    main()




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Set image_center in CXI file
Author: Sergey Bobkov
"""

import os
import sys
import shutil
import subprocess
import argparse
import numpy as np
from tqdm import tqdm

from spi_processing import center, cxidata


def main():
    parser = argparse.ArgumentParser(description='Set image center in CXI data')
    parser.add_argument('files', metavar='FILE', nargs='+', help='Input files')
    parser.add_argument('-o', '--out', dest='output_dir', help="Output directory")
    parser.add_argument('-x', dest='center_x', type=float, help="x coordinate (pixels)")
    parser.add_argument('-y', dest='center_y', type=float, help="y coordinate (pixels)")
    parser.add_argument('-z', dest='center_z', type=float, default=0, help="z coordinate (pixels)")
    parser.add_argument('-i', '--input', dest='input', metavar='FILE',
                        help='Copy beam position from CXI file')
    parser_args = parser.parse_args()

    input_files = parser_args.files
    output_dir = parser_args.output_dir
    center_x = parser_args.center_x
    center_y = parser_args.center_y
    center_z = parser_args.center_z
    input_file = parser_args.input

    for fname in input_files:
        if not os.path.isfile(fname):
            parser.error("File {} doesn't exist".format(fname))

    if output_dir:
        os.makedirs(output_dir, exist_ok=True)
        sys.stderr.write("Copying to output\n")
        os.makedirs(output_dir, exist_ok=True)
        new_input_files = []
        for fname in tqdm(input_files):
            new_fname = os.path.join(output_dir, os.path.basename(fname))
            shutil.copy(fname, new_fname)
            subprocess.call(["/bin/chmod", "u+w", new_fname])
            new_input_files.append(new_fname)
        input_files = new_input_files

    if (input_file is None) == ((center_x is None) or (center_y is None)):
        parser.error("Define center position (x and y) or input CXI file")

    if (center_x is None) != (center_y is None):
        parser.error("Define both x and y positions or input CXI file")

    if input_file is not None:
        image_ids = cxidata.get_image_groups(input_file)
        names = cxidata.get_names(input_file, image_ids[0])
        if 'image_center' not in names:
            parser.error("Input CXI file miss image_center dataset")
        new_center = cxidata.read_dataset(input_file, image_ids[0], 'image_center')
    else:
        new_center = np.array([center_x, center_y, center_z])

    for fname in tqdm(input_files):
        center.save(fname, new_center)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Correct background scattering in diffraction data in CXI files
Author: Sergey Bobkov
"""

import os
import sys
import shutil
import subprocess
import argparse
from tqdm import tqdm

from spi_processing import correct_background


def main():
    parser = argparse.ArgumentParser(
        description='Correct background scattering in diffraction data in CXI files')
    parser.add_argument('files', metavar='FILE', nargs='+', help='Input files')
    parser.add_argument('-o', '--out', dest='output_dir', help="Output directory")
    parser.add_argument('-M', '--max', default=50, help="Maximum background intensity")
    parser.add_argument('-O', '--overflow', default=70,
                        help="Maximum intensity when detector is still linear")
    parser.add_argument('-r', '--report', default="correct_background_report.pdf",
                        help="Report file")
    parser.add_argument('--no-correct', dest='nocorr', action='store_true',
                        help="Do not correct data, generate report for already corrected data")
    parser_args = parser.parse_args()

    input_files = parser_args.files
    output_dir = parser_args.output_dir
    max_background = parser_args.max
    overflow_level = parser_args.overflow
    report_path = parser_args.report
    nocorr = parser_args.nocorr

    for fname in input_files:
        if not os.path.isfile(fname):
            parser.error("File {} doesn't exist".format(fname))

    old_mean_list = []
    new_mean_list = []

    if nocorr:
        if not output_dir:
            parser.error("--no-correct require OUTPUT_DIR")

        new_input_files = []
        for fname in input_files:
            new_fname = os.path.join(output_dir, os.path.basename(fname))
            new_input_files.append(new_fname)
            if not os.path.isfile(new_fname):
                parser.error("Output file {} doesn't exist".format(fname))

        sys.stderr.write("Reading files\n")
        for i in tqdm(range(len(input_files))):
            old_mean_list.append(correct_background.compute_mean_data(input_files[i]))
            new_mean_list.append(correct_background.compute_mean_data(new_input_files[i]))
    else:
        if output_dir:
            os.makedirs(output_dir, exist_ok=True)
            sys.stderr.write("Copying to output\n")
            os.makedirs(output_dir, exist_ok=True)
            new_input_files = []
            for fname in tqdm(input_files):
                new_fname = os.path.join(output_dir, os.path.basename(fname))
                shutil.copy(fname, new_fname)
                subprocess.call(["/bin/chmod", "u+w", new_fname])
                new_input_files.append(new_fname)
            input_files = new_input_files

        sys.stderr.write("Processing files\n")
        for fname in tqdm(input_files):
            old_mean_list.append(correct_background.compute_mean_data(fname))
            correct_background.correct_background(fname, max_background, overflow_level)
            new_mean_list.append(correct_background.compute_mean_data(fname))

    sys.stderr.write("Generating report...")
    correct_background.generate_report(input_files, old_mean_list, new_mean_list, report_path)
    sys.stderr.write("Done\n")


if __name__ == '__main__':
    main()



from setuptools import setup

def readme():
    with open('README.md') as f:
        return f.read()

setup(name='spi_processing',
      version='0.1',
      description='Module for processing of diffraction images in CXI format.',
      long_description=readme(),
      long_description_content_type="text/markdown",
      url='https://gitlab.com/spi_xfel/spi_processing',
      author='Sergey Bobkov',
      author_email='s.bobkov@grid.kiae.ru',
      license='MIT',
      python_requires='>=3.6',
      install_requires=['numba',
                        'numpy',
                        'scipy',
                        'matplotlib',
                        'pandas',
                        'h5py',
                        'tqdm'],
      packages=['spi_processing'],
      scripts=['scripts/spi_combine.py',
               'scripts/spi_compute_photons.py',
               'scripts/spi_correct_background.py',
               'scripts/spi_estimate_center.py',
               'scripts/spi_estimate_size.py',
               'scripts/spi_filter.py',
               'scripts/spi_plot_histogram.py',
               'scripts/spi_refine_center.py',
               'scripts/spi_set_center.py'],
      zip_safe=False)



"""
Scripts for estimation of image center for diffraction images in CXI file
Author: Sergey Bobkov
"""

from typing import Tuple, Union
import numpy as np
from scipy.optimize import curve_fit


def simulate_pattern(shape: Tuple[int, int], center: np.ndarray,
                     sphere_r: float, scale: float, freq_step: float):
    """Construct 2d scattering pattern for sphere

    Keyword arguments:
    shape -- shape and mask of pattern to be constructed
    center -- position of scattering center (pixels)
    sphere_r -- mid radius of a sphere
    scale -- intensity value for pattern
    freq_step -- frequency increment for 1 pixel, equal to pixel_size/(wavelength*distance)

    Return:
    result -- 2d array with the simulated sphere scattering
    """
    y_range = np.arange(shape[0], dtype=np.float64)
    x_range = np.arange(shape[1], dtype=np.float64)

    x_grid, y_grid = np.meshgrid(x_range, y_range)
    x_grid -= center[0]
    y_grid -= center[1]

    # r_grid = np.round(np.sqrt(x_grid**2 + y_grid**2))
    # r_grid[mask_data != 0] = 0
    # vals, indices = np.unique(r_grid, return_inverse=True)
    # return sphere_form_factor(vals*one_pix_freq, sphere_mid, scale)[indices].reshape(r_grid.shape)

    r_grid = np.sqrt(x_grid**2 + y_grid**2)
    return form_factor(r_grid*freq_step, sphere_r, scale)


def refine_size(psd_data: np.ndarray, size_start: float, freq_step: float,
                r_min: int = 0, r_max: int = 0, return_fit: bool = False):
    """Refine particle size by fit of PSD data with spherical form factor
    Do not search for global minimum, find closest local minimum

    Keyword arguments:
    psd_data -- input PSD data
    size_start -- start size
    freq_step -- frequency increment for 1 pixex, equal to pixel_size/(wavelength*distance)
    r_min -- minimum radius (index of PSD array) used to fit
    r_max -- maximum radius (index of PSD array) used to fit

    Return:
    result -- dict with estimation results, 'r_mid' - radius of sphere with best fit,
              'size' - predicted size of a particle,
              'scale' - arbitrary intensity of a beam
    """
    if r_min == 0:
        beamstop = int(np.argmax(psd_data > 0))
        r_min = beamstop + 5
    if r_max == 0:
        r_max = len(psd_data)

    frequency_coord = np.arange(len(psd_data))*freq_step
    data_to_fit = psd_data[r_min:r_max]
    scale_start = data_to_fit.mean()/ \
                  form_factor(frequency_coord[r_min:r_max], size_start/2, 1).mean()

    pstart = [size_start/2, scale_start]

    try:
        popt, _ = curve_fit(form_factor,
                            frequency_coord[r_min:r_max],
                            data_to_fit,
                            p0=pstart, bounds=(0, np.inf))
    except RuntimeError:
        popt = pstart

    r_mid = popt[0]
    scale = popt[1]
    size = 2*r_mid
    result = {'r_mid':r_mid, 'size':size, 'scale':scale}

    if return_fit:
        radial_fit = np.array([form_factor(x, r_mid, scale) for x in frequency_coord])
        result['fit'] = radial_fit
        result['fit_start'] = r_min
        result['fit_end'] = r_max

    return result


def compute_size(first_min: int, freq_step: float):
    """Compute size of a sphere by position of first min in sphere form factor

    Keyword arguments:
    first_min_pix -- position of first minimum of form factor (pix)
    freq_step -- Spatial frequency for angle of one detector pixel (1/Angstrom)

    Return:
    size -- sphere size (pixel_size/(wavelength*distance)) (Angstrom)
    """

    return 4.5 / (np.pi*first_min*freq_step)


def form_factor(x_val: Union[float, np.ndarray], sphere_r: float, scale: float):
    """Calculate sphere form factor

    Keyword arguments:
    x -- spatial frequency
    r -- sphere radius
    scale -- intensity multiplier

    Return:
    res -- sphere form factor value
    """

    q_by_r = x_val*2*np.pi*sphere_r
    vol = (4/3)*np.pi*(sphere_r**3)
    with np.errstate(divide='ignore', invalid='ignore'):
        res = (scale/vol)*((3*vol*(np.sin(q_by_r) - q_by_r*np.cos(q_by_r))/q_by_r**3)**2)
    return np.nan_to_num(res)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Estimate particle size for diffraction images by fitting of PSD with spherical form-factor
Author: Sergey Bobkov
"""

import os
import sys
import shutil
import subprocess
import argparse
import numpy as np
from tqdm import tqdm

from spi_processing import particle_size


def main():
    parser = argparse.ArgumentParser(description='Estimate particle size for diffraction images by fitting of PSD with spherical form-factor')
    parser.add_argument('files', metavar='FILE', nargs='+', help='Input files')
    parser.add_argument('-o', '--out', dest='output_dir', help="Output directory")
    parser.add_argument('-p', '--percentage', type=float, default=1,
                        help="Percentage of used angular data, 1 for all, 0.1 for brightest 10 percent")
    parser.add_argument('-m', '--s_min', type=int, default=0,
                        help="Minimum possible size (Angstrom)")
    parser.add_argument('-M', '--s_max', type=int, default=0,
                        help="Maximum possible size (Angstrom)")
    parser.add_argument('-r', '--r_min', type=int, default=0,
                        help="Minimal fit radius (Pixels)")
    parser.add_argument('-R', '--r_max', type=int, default=0,
                        help="Maximum fit radius (Pixels)")
    parser.add_argument('-w', '--wavelength', type=float, required=True,
                        help="Radiation wavelength (Angstrom)")
    parser.add_argument('-d', '--distance', type=float, required=True,
                        help="Detector distance (meters)")
    parser.add_argument('--pix', dest='pixel', type=float, required=True,
                        help="Pixel size (meters)")
    parser.add_argument('-i','--interp', type=int, default=5,
                        help="Interpolation factor")
    parser.add_argument('-n', '--nsize', type=int, default=300,
                        help="Number of sizes tested within size range")
    parser.add_argument('--icosahedron', action='store_true',
                        help="Size of particle will be computed with assumption that it is a regular icosahedron")
    parser_args = parser.parse_args()

    input_files = parser_args.files
    output_dir = parser_args.output_dir
    percentage = parser_args.percentage
    size_min = parser_args.s_min
    size_max = parser_args.s_max
    r_min = parser_args.r_min
    r_max = parser_args.r_max
    wavelength = parser_args.wavelength
    distance = parser_args.distance
    pixel_size = parser_args.pixel
    interp = parser_args.interp
    nsize = parser_args.nsize
    icosahedron = parser_args.icosahedron

    for fname in input_files:
        if not os.path.isfile(fname):
            parser.error("File {} doesn't exist".format(fname))

    if icosahedron:
        # r_mid = a/4 * (1 + sqrt(5))
        # r_outer = a/4 * sqrt(10 + 2*sqrt(5))
        size_multiplier = np.sqrt(10 + 2*np.sqrt(5)) / (1 + np.sqrt(5))
        if size_max is None:
            size_max /= size_multiplier
        if size_min is None:
            size_min /= size_multiplier

    if output_dir:
        os.makedirs(output_dir, exist_ok=True)
        sys.stderr.write("Copying to output\n")
        os.makedirs(output_dir, exist_ok=True)
        new_input_files = []
        for fname in tqdm(input_files):
            new_fname = os.path.join(output_dir, os.path.basename(fname))
            shutil.copy(fname, new_fname)
            subprocess.call(["/bin/chmod", "u+w", new_fname])
            new_input_files.append(new_fname)
        input_files = new_input_files

    sys.stderr.write("Processing files\n")
    for fname in tqdm(input_files):
        particle_size.compute_psd_data(fname, r_max, percentage)
        particle_size.estimate_size(fname, wavelength, distance, pixel_size, r_min, r_max, interp,
                                    size_min, size_max, nsize)

        if icosahedron:
            particle_size.correct_size(fname, size_multiplier)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Compute number of photons and lit-pixels for data in cxi dataset
Author: Sergey Bobkov
"""

import os
import sys
import shutil
import subprocess
import argparse
from tqdm import tqdm

from spi_processing.compute_photons import compute_photons


def main():
    parser = argparse.ArgumentParser(
        description="Compute number of photons and litpixels in CXI data")
    parser.add_argument('files', metavar='FILE', nargs='+', help='Input files')
    parser.add_argument('-o', '--out', dest='output_dir', help="Output directory")
    parser_args = parser.parse_args()

    input_files = parser_args.files
    output_dir = parser_args.output_dir

    for fname in input_files:
        if not os.path.isfile(fname):
            parser.error("File {} doesn't exist".format(fname))

    if output_dir:
        sys.stderr.write("Copying to output\n")
        os.makedirs(output_dir, exist_ok=True)
        new_input_files = []
        for fname in tqdm(input_files):
            new_fname = os.path.join(output_dir, os.path.basename(fname))
            shutil.copy(fname, new_fname)
            subprocess.call(["/bin/chmod", "u+w", new_fname])
            new_input_files.append(new_fname)
        input_files = new_input_files

    sys.stderr.write("Processing files\n")
    for fname in tqdm(input_files):
        compute_photons(fname)


if __name__ == '__main__':
    main()
